package cz.fidentis.analyst.visitors.mesh;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import java.util.List;
import javax.vecmath.Vector3d;

/**
 * Calculates maximum or minimum curvatures of mesh vertices using the combination
 * of Gaussian and mean curvatures.
 * 
 * @author Natalia Bebjakova
 * @author Radek Oslejsek
 */
public class DeltaCurvature extends GaussianCurvature {
    
    private boolean maxCurvature = true;
    
    /**
     * Constructor.
     * @param max If {@code true}, then maximum curvature is return. 
     * Otherwise, minimum curvature is computed.
     */
    public DeltaCurvature(boolean max) {
        this.maxCurvature = max;
    }
    
    @Override
    protected double calculateCurvature(MeshFacet facet, List<CurvTriangle> triangles, int vertA) {
        List<Integer> neighbouringTriangles = facet.getCornerTable().getTriangleIndexesByVertexIndex(vertA);
        
        if (neighbouringTriangles.isEmpty()) {
            return Double.NaN;
        }
        
        double sumArea = 0.0;
        double sumAngles = 0.0;
        Vector3d pointSum = new Vector3d();
        
        // for all surrounding triangles:
        for (int i = 0; i < neighbouringTriangles.size(); i++) {
            CurvTriangle ctri = triangles.get(neighbouringTriangles.get(i));
            CurvTriangle tNext = triangles.get(neighbouringTriangles.get((i + 1) % neighbouringTriangles.size()));
            
            sumArea += computeArea(ctri, facet.getVertex(vertA));
            sumAngles +=  ctri.alpha(facet.getVertex(vertA));
            
            Vector3d aux = new Vector3d(ctri.vertC(facet.getVertex(vertA)));
            aux.sub(ctri.vertA(facet.getVertex(vertA)));
            aux.scale(ctri.betaCotan(facet.getVertex(vertA)) + tNext.gammaCotan(facet.getVertex(vertA)));
            pointSum.add(aux);
        }
        
        double gaussian = (2.0 * Math.PI - sumAngles) / sumArea;
        double mean = 0.25 * sumArea * pointSum.length();
        double delta = Math.max(0, Math.pow(mean, 2) - gaussian);
        
        return (maxCurvature) ? mean + Math.sqrt(delta) : mean - Math.sqrt(delta);
        
        
        /*
        double gaussianCurvature = super.calculateCurvature(facet, triangles, centerIndex);
        if (Double.isNaN(gaussianCurvature)) {
            return Double.NaN;
        }
        double meanCurvature = getMeanCurvature(facet, triangles, centerIndex);
        double meanCurvatureSqr = meanCurvature * meanCurvature;
        if (meanCurvatureSqr <= gaussianCurvature) { 
            return meanCurvature;
        }
        return meanCurvature + Math.sqrt(meanCurvatureSqr - gaussianCurvature);
        */
    }
    
    /*
    private Vector3d computeLaplaceBeltrami(MeshFacet facet, List<MeshTriangle> triangles, int centerIndex) {
        List<Integer> neighbouringTriangles = facet.getCornerTable().getTriangleIndexesByVertexIndex(centerIndex);

        if (neighbouringTriangles.isEmpty()) {
            return new Vector3d();
        }
        
        double areaSum = 0;
        Vector3d pointSum = new Vector3d();
        List<Vector3d> voronoiPoints = facet.calculateVoronoiPoints();
        
        for (int i = 0; i < neighbouringTriangles.size(); i++) {
            MeshTriangle t = triangles.get(neighbouringTriangles.get(i));
            MeshTriangle tNext = triangles.get(neighbouringTriangles.get((i + 1) % neighbouringTriangles.size()));
            double area = 0;
            
            MeshPoint centerPoint = facet.getVertices().get(centerIndex);
            Vector3d voronoiPoint = voronoiPoints.get(neighbouringTriangles.get(i));

            Vector3d aux = null;
            if (t.vertex1 == facet.getVertex(centerIndex)) {
                area = voronoiPoint.x;
                aux = new Vector3d(t.vertex2.getPosition());
            } else if (t.vertex2 == facet.getVertex(centerIndex)) {
                area = voronoiPoint.y;
                aux = new Vector3d(t.vertex3.getPosition());
            } else if (t.vertex3 == facet.getVertex(centerIndex)) {
                area = voronoiPoint.z;
                aux = new Vector3d(t.vertex1.getPosition());
            }
            aux.sub(centerPoint.getPosition());
            aux.scale(cotAlfa(centerPoint, t) + cotBeta(centerPoint, tNext));
            pointSum.add(aux);
            areaSum += area;
        }
        pointSum.scale(1.0 / (2.0 * areaSum));
        return pointSum;
    }
    */
    
    /**
     * Calculates angle of the triangle according to center point
     * TODO: optimize @see https://stackoverflow.com/questions/31159016/how-to-efficiently-calculate-cotangents-from-vectors
     * 
     * @param centerPoint center point
     * @param t triangle
     * @return angle
     */
    /*
    private double cotAlfa(MeshPoint centerPoint, MeshTriangle t) {        
        double alfaCos = 0;
        
        if (t.vertex1 == centerPoint) {
            MeshPoint e1 = t.vertex1.subtractPosition(t.vertex3);
            e1 = e1.dividePosition(e1.abs());
            
            MeshPoint e2 = t.vertex2.subtractPosition(t.vertex3);
            e2 = e2.dividePosition(e2.abs());
            
            alfaCos = e1.dotProduct(e2);
        } else if (t.vertex2 == centerPoint) {
            MeshPoint e1 = t.vertex2.subtractPosition(t.vertex1);
            e1 = e1.dividePosition(e1.abs());
            
            MeshPoint e2 = t.vertex3.subtractPosition(t.vertex1);
            e2 = e2.dividePosition(e2.abs());
            
            alfaCos = e1.dotProduct(e2);
        } else if (t.vertex3 == centerPoint){
            MeshPoint e1 = t.vertex3.subtractPosition(t.vertex2);
            e1 = e1.dividePosition(e1.abs());
                
            MeshPoint e2 = t.vertex1.subtractPosition(t.vertex2);
            e2 = e2.dividePosition(e2.abs());
            
            alfaCos = e1.dotProduct(e2);
        }
        double alfa = Math.acos(alfaCos);
        return 1 / Math.tan(alfa); // cotangent alpha
    }
    */
    
    /**
     * Calculates angle of the triangle according to center point
     * TODO: optimize @see https://stackoverflow.com/questions/31159016/how-to-efficiently-calculate-cotangents-from-vectors
     * 
     * @param centerPoint center point
     * @param t triangle
     * @return angle
     */
    /*
    private double cotBeta(MeshPoint centerPoint, MeshTriangle t) {
        double betaCos = 0;
            
        if (t.vertex1 == centerPoint) {
            MeshPoint e1 = t.vertex1.subtractPosition(t.vertex2);
            e1 = e1.dividePosition(e1.abs());
            
            MeshPoint e2 = t.vertex3.subtractPosition(t.vertex2);
            e2 = e2.dividePosition(e2.abs());
            
            betaCos = e1.dotProduct(e2);
        } else if (t.vertex2 == centerPoint) {
            MeshPoint e1 = t.vertex2.subtractPosition(t.vertex3);
            e1 = e1.dividePosition(e1.abs());
            
            MeshPoint e2 = t.vertex1.subtractPosition(t.vertex3);
            e2 = e2.dividePosition(e2.abs());
            
            betaCos = e1.dotProduct(e2);
        } else if (t.vertex3 == centerPoint){
            MeshPoint e1 = t.vertex3.subtractPosition(t.vertex1);
            e1 = e1.dividePosition(e1.abs());
                
            MeshPoint e2 = t.vertex2.subtractPosition(t.vertex1);
            e2 = e2.dividePosition(e2.abs());
            
            betaCos = e1.dotProduct(e2);
        }
            double beta = Math.acos(betaCos);
            return 1 / Math.tan(beta); // cotangent beta
    }
    */
}
